import numpy as np
import pandas as pd
from pytest import raises

from colassigner import ChildColAssigner, ColAssigner, get_all_cols


def test_flat_assigner():
    class Cols(ColAssigner):
        def col_a(self, df):
            return df.iloc[:, -1] * 2

        def acol(self, df):
            return df.loc[:, Cols.col_a] - 1

    assert Cols.col_a == "col_a"
    assert Cols.acol == "acol"
    assert get_all_cols(Cols) == ["acol", "col_a"]  # alphabetical

    df = pd.DataFrame({"a": [1, 2, 3]}).pipe(Cols())
    assert df.loc[:, Cols.col_a].tolist() == [2, 4, 6]
    assert df.loc[:, Cols.acol].tolist() == [1, 3, 5]


def test_inited_assigner():
    class Cols(ColAssigner):
        def __init__(self, n) -> None:
            self.n = n

        def col(self, df):
            return (df < self.n).all(axis=1)

    df = pd.DataFrame({"a": [1, 2, 3], "b": [10, 0, 1]}).pipe(Cols(3))
    assert df.loc[:, Cols.col].tolist() == [False, True, False]
    assert get_all_cols(Cols) == ["col"]


def test_wrong_colname():
    with raises(ValueError):

        class ColW(ColAssigner):
            def mro(self, df):
                return 4  # pragma: nocover


def test_child_assigner():
    class AssChild(ChildColAssigner):
        def nothing(self):
            return False

    class Cols(ColAssigner):
        def col_one(self, df):
            return 1

        class SubCol(ColAssigner):
            def fing(self, df):
                return df.sum(axis=1)

            class SubSubCol(ColAssigner):
                _prefix = "pref_"

                def sub_x(self, df):
                    return 0

                def sub_y(self, df):
                    return self._prefix + df[Cols.col_one].astype(str)

            class SubSubCol2(SubSubCol):
                _prefix = "pref2_"

        sep_child = AssChild

    assert Cols.SubCol.fing == "sub_col__fing"
    assert Cols.SubCol.SubSubCol.sub_x == "sub_col__sub_sub_col__sub_x"
    assert Cols.SubCol.SubSubCol2.sub_y == "sub_col__sub_sub_col_2__sub_y"
    assert get_all_cols(Cols) == [
        "sub_col__sub_sub_col__sub_x",
        "sub_col__sub_sub_col__sub_y",
        "sub_col__sub_sub_col_2__sub_x",
        "sub_col__sub_sub_col_2__sub_y",
        "sub_col__fing",
        "col_one",
        "sep_child__nothing",
    ]

    df = pd.DataFrame({"a": [1, 2, 3]}).pipe(Cols())
    assert (df.loc[:, Cols.SubCol.fing] == (df["a"] + df[Cols.col_one])).all()
    assert (df[Cols.SubCol.SubSubCol2.sub_y] == "pref2_1").all()


def test_subassigner_data_reuse():
    class IntSides(ChildColAssigner):
        def __init__(self, df, parent_assigner: "GbReindex") -> None:
            self.arr = parent_assigner.arr

        def lower(self):
            return np.floor(self.arr).astype(int)

        def upper(self):
            return np.ceil(self.arr).astype(int)

    class GbReindex(ChildColAssigner):
        main_col = ...

        def __init__(self, df, bc: "BaseCols"):
            self.arr = bc.base_gb.reindex(df[self.main_col]).values

        def values(self):
            return self.arr

        sides = IntSides

    class BaseCols(ColAssigner):
        def __init__(self, base_df):
            self.base_gb = base_df.groupby("cat")["num"].mean()

        class GbB(GbReindex):
            main_col = "b"

        class GbC(GbReindex):
            main_col = "c"

        def prod(self, df):
            return df.loc[:, [BaseCols.GbB.sides.lower, BaseCols.GbC.values]].prod(
                axis=1
            )

    df1 = pd.DataFrame({"cat": ["x", "y", "y"], "num": [2, 3, 4]})
    # add mean of category, with lower and upper bounds
    # for 2 different columns, add product of two of these
    df2 = pd.DataFrame({"b": ["x", "y", "x"], "c": ["y", "y", "x"]}).pipe(BaseCols(df1))

    assert df2.loc[:, BaseCols.GbB.values].tolist() == [2, 3.5, 2]
    assert df2.loc[:, BaseCols.GbC.sides.lower].tolist() == [3, 3, 2]
    assert df2.loc[:, BaseCols.GbC.sides.upper].tolist() == [4, 4, 2]
    assert df2.loc[:, BaseCols.prod].tolist() == [7, 10.5, 4]



import re

wswitch_rex = re.compile("(?=([a-z|0-9|A-Z][A-Z|0-9]))")
nonswitch_rex = re.compile("[0-9]{2}")


def camel_to_snake(cc_str):
    out = cc_str
    for group in wswitch_rex.findall(cc_str):
        if nonswitch_rex.findall(group):
            continue
        out = out.replace(group, "_".join(group))
    return out.lower()



import numpy as np

_T_COL = "__treated"


def experiment(assigner_inst, base_df, col_to_treat, treat_rate=0.2, seed=742):
    treat_arr = np.random.RandomState(seed).rand(base_df.shape[0]) < treat_rate
    orig_fun = getattr(assigner_inst, col_to_treat)

    def f(df, *args, **kwargs):
        out = orig_fun(df, *args, **kwargs)
        return out + df[_T_COL].astype(int)

    f.__name__ = col_to_treat
    _reset_atts(assigner_inst, col_to_treat, f)
    out = base_df.assign(**{_T_COL: treat_arr}).pipe(assigner_inst)
    _reset_atts(assigner_inst, col_to_treat, orig_fun)
    return out


def measure_effect(assigner, base_df, cause_col, effect_col, treat_rate=0.2, seed=742):
    exp_df = experiment(assigner, base_df, cause_col, treat_rate, seed)
    return exp_df.groupby(_T_COL)[effect_col].mean().pipe(lambda s: s[True] - s[False])


def _reset_atts(assigner_inst, reset_att, reset_val):
    # needed to preserve order
    for c in assigner_inst.__dir__():
        if c.startswith("_"):
            continue
        v = reset_val if c == reset_att else getattr(assigner_inst, c)
        setattr(assigner_inst, c, v)



from abc import ABCMeta

from colassigner.util import camel_to_snake

from .constants import DEFAULT_PP, FORBIDDEN_NAMES, PREFIX_SEP
from .type_hinting import get_return_hint


class ColMeta(ABCMeta):
    def __new__(cls, name, bases, local):
        for attr in local:
            if (attr in FORBIDDEN_NAMES) or (
                PREFIX_SEP in attr and not attr.startswith("_")
            ):
                raise ValueError(
                    f"Column name can't be either {FORBIDDEN_NAMES}. "
                    f"And can't contain the string {PREFIX_SEP}. "
                    f"{attr} is given"
                )
        return super().__new__(cls, name, bases, local)

    def __init__(self, name: str, bases, namespace) -> None:
        super().__init__(name, bases, namespace)
        self._parent_prefixes = namespace.get("_parent_prefixes", DEFAULT_PP)

    def __getattribute__(cls, attid):
        "so that Cls.xy returns a string for column access"

        att_value = super().__getattribute__(attid)
        if attid.startswith("_") or (attid in FORBIDDEN_NAMES):
            return att_value

        new_pref_arr = (*cls._parent_prefixes, camel_to_snake(attid))

        if isinstance(att_value, ColMeta):

            class _C(att_value):
                _parent_prefixes = new_pref_arr

            return _C

        return_hint_str = get_hint_str(att_value)

        return PREFIX_SEP.join(filter(None, (*new_pref_arr, return_hint_str)))

    def __repr__(cls) -> str:
        base = super().__repr__()
        return base + f" ({', '.join(cls.__col_dir__())})"

    def __getcoltype__(cls, attid):
        colval = super().__getattribute__(attid.split(PREFIX_SEP)[-1])
        return colval

    def __col_dir__(cls):
        return [k for k in dir(cls) if not k.startswith("_")]


def get_all_cols(cls: ColMeta):
    """returns a list of strings of all columns given by the type

    can also be used for nested structues of columns
    """
    out = []
    for attid in cls.__col_dir__():
        attval = getattr(cls, attid)
        if isinstance(attval, ColMeta):
            out += get_all_cols(attval)
            continue
        if isinstance(cls, ColMeta):
            out.append(attval)
    return out


def get_att_value(accessor: ColMeta, attname: str):
    """get the true assigned value for the class attribute"""
    return accessor.__getcoltype__(attname)


def get_hint_str(val):
    return_hint = get_return_hint(val)
    if isinstance(return_hint, ColMeta):
        cols = get_all_cols(return_hint)
        if len(cols) != 1:
            raise ValueError(f"cols in {return_hint} is not 1 : {cols}")
        return cols[0]
    if isinstance(return_hint, str):
        return return_hint



from .constants import PREFIX_SEP
from .meta_base import ColMeta
from .util import camel_to_snake


class ColAccessor(metaclass=ColMeta):
    """describe and access raw columns

    useful for
    - getting column names from static analysis
    - documenting types
    - dry describing nested structures

    e. g.

    >>> class LocationCols(ColAccessor):
    ...     lon = float
    ...     lat = float

    >>> class TableCols(ColAccessor):
    ...     col1 = int
    ...     col2 = str
    ...     foreign_key1 = "name_of_key"
    ...
    ...     class NestedCols(ColAccessor):
    ...         s = str
    ...         x = float
    ...
    ...     start_loc = LocationCols
    ...     end_loc = LocationCols

    >>> TableCols.start_loc.lat
    'start_loc__lat'

    """


class ColAssigner(ColAccessor):
    """define functions that create columns in a dataframe

    later the class attributes can be used to access the column
    can be used to created nested structures of columns

    either by assigning or inheriting within:

    class MyStaticChildAssigner(ColAssigner):

        pass

    class MyAssigner(ColAssigner):

        class MySubAssigner(ColAssigner):
            pass

        chass1 = MyStaticChildAssigner
    """

    def __call__(self, df, carried_prefixes=()):
        # dir() is alphabetised object.__dir__ is not
        # important here if assigned cols rely on each other
        for attid in self.__dir__():
            if attid.startswith("_"):
                continue
            att = getattr(self, attid)
            new_pref_arr = (*carried_prefixes, camel_to_snake(attid))
            if isinstance(att, ColMeta):
                if ChildColAssigner in att.mro():
                    inst = att(df, self)
                else:
                    inst = att()
                df = inst(df, carried_prefixes=new_pref_arr)
            elif callable(att):
                col_name = getattr(type(self), attid)
                colname = PREFIX_SEP.join((*carried_prefixes, col_name))
                df = df.assign(**{colname: self._call_att(att, df)})
        return df

    @staticmethod
    def _call_att(att, df):
        return att(df)


class ChildColAssigner(ColAssigner):
    """assigner specifically for nested structures

    methods of these are not called with parameters

    the dataframe and the parent assigner are passed
    to the __init__ method as parameters
    """

    def __init__(self, df, parent_assigner: ColAssigner) -> None:
        pass

    @staticmethod
    def _call_att(att, _):
        return att()



"""Helper for assigning and accessing pandas columns"""

# flake8: noqa
from .core import ChildColAssigner, ColAccessor, ColAssigner
from .dag_from_ast import get_dag
from .experiment import experiment, measure_effect
from .meta_base import get_all_cols, get_att_value
from .type_hinting import Col
from .util import camel_to_snake

__version__ = "0.4.0"



from typing import ForwardRef, Generic, TypeVar

T = TypeVar("T")

ERR_TXT = "return type hint not wrapped in Col[...]"


class Col(Generic[T]):
    pass


def get_hint_col_type(rethint):
    assert is_type_hint_origin(rethint, Col), ERR_TXT
    arg = rethint.__args__[0]
    if isinstance(arg, ForwardRef):
        return arg.__forward_arg__
    return arg


def is_type_hint_origin(hint, cls):
    try:
        return hint.__origin__ is cls
    except AttributeError:
        return False


def get_return_hint(fun):
    ret_annotation = getattr(fun, "__annotations__", {}).get("return")
    if ret_annotation is None:
        return None
    return get_hint_col_type(ret_annotation)



from colassigner import ColAccessor, get_att_value


def test_type_retention():
    class TestCols(ColAccessor):

        fing = int

        class SubCol(ColAccessor):
            a = int
            b = float

            class SubSubCol(ColAccessor):
                x = str
                y = str

            class SubSubCol2(SubSubCol):
                pass

    assert get_att_value(TestCols, TestCols.fing) == int
    assert get_att_value(TestCols.SubCol, TestCols.SubCol.b) == float
    assert (
        get_att_value(TestCols.SubCol.SubSubCol2, TestCols.SubCol.SubSubCol2.x) == str
    )



from colassigner.util import camel_to_snake


def test_cc_to_snake():
    s = "ClassToBeRenamed"
    s2 = "ClassWithAOneLetterWord"
    s3 = "ClassWith1Number"
    s4 = "ClassWith10Number"

    assert camel_to_snake(s) == "class_to_be_renamed"
    assert camel_to_snake(s2) == "class_with_a_one_letter_word"
    assert camel_to_snake(s3) == "class_with_1_number"
    assert camel_to_snake(s4) == "class_with_10_number"



import numpy as np
import pandas as pd

from colassigner import ColAssigner, experiment, measure_effect
from colassigner.experiment import _T_COL


class Cols(ColAssigner):
    def __init__(self) -> None:
        self._rng = None

    def col_x(self, df):
        self._rng = np.random.RandomState(123)
        return self._rng.random(df.shape[0])

    def col_b(self, df):
        return df[Cols.col_x] * 2

    def col_c(self, df):
        return df[Cols.col_x] * 0.5

    def col_d(self, df):
        return df[Cols.col_b] - df[Cols.col_c]


def test_effect():
    cass = Cols()
    df = pd.DataFrame(index=range(1000)).pipe(cass)
    eff = measure_effect(cass, df, Cols.col_x, Cols.col_d)
    assert round(eff, 1) == 1.5


def test_experiment():
    cass = Cols()
    edf = experiment(cass, pd.DataFrame(index=range(500)), Cols.col_b, 0.3)
    treat_means = edf.groupby(_T_COL).mean().round(1).loc[True, :].tolist()
    assert treat_means == [0.5, 2.0, 0.2, 1.7]



from colassigner import ColAssigner, get_dag


class Cols(ColAssigner):  # pragma: no cover
    # TODO: nested

    def __init__(self):
        self._a = 4

    def col_a(self, df):
        return range(self._a, df.shape[0] + self._a)

    def col_b(self, df):
        return df.loc[:, Cols.col_a] - 1

    def col_c(self, df):
        return self._helper(df, 3)

    def col_d(self, df):
        return self._helper2(df, Cols.col_c)

    def col_e(self, df):
        return self._helper2(df, col=Cols.col_a) + self._helper3(df)

    def _helper(self, df, n):
        if n > 100:
            return self._helper(df, n / 2)
        return df.loc[:, Cols.col_b].pipe(lambda s: s * 3)

    def _helper2(self, df, col):
        for _ in range(3):
            pass
        return df[col] ** 2

    def _helper3(self, df):
        try:
            "boo"
        except ValueError:
            raise IndexError()
        return self._helper2(df, Cols.col_d) - self._helper(df, 3)


true_dag = [
    ("col_a", "col_b"),
    ("col_b", "col_c"),
    ("col_c", "col_d"),
    ("col_a", "col_e"),
    ("col_b", "col_e"),
    ("col_d", "col_e"),
]


def test_graph_detection():
    assert sorted(true_dag) == sorted(get_dag(Cols))



from colassigner import ColAccessor, get_all_cols


def test_flat_accessor():
    class Cols(ColAccessor):
        x = int
        a = float

    assert Cols.x == "x"
    assert Cols.a == "a"


def test_nested_accessor():
    class GrandChildCols(ColAccessor):
        x = str
        y = str

    class ChildCols(ColAccessor):
        a = int
        b = float
        grandchild_a = GrandChildCols
        grandchild_b = GrandChildCols

    class Cols(ColAccessor):

        fing = int
        assigned_child = ChildCols

        class InheritedChild(ChildCols):
            pass

    assert Cols.fing == "fing"
    assert Cols.assigned_child.a == "assigned_child__a"
    assert Cols.assigned_child.grandchild_a.x == "assigned_child__grandchild_a__x"
    assert Cols.assigned_child.grandchild_b.y == "assigned_child__grandchild_b__y"
    assert Cols.InheritedChild.b == "inherited_child__b"

    assert get_all_cols(Cols) == [
        "inherited_child__a",
        "inherited_child__b",
        "inherited_child__grandchild_a__x",
        "inherited_child__grandchild_a__y",
        "inherited_child__grandchild_b__x",
        "inherited_child__grandchild_b__y",
        "assigned_child__a",
        "assigned_child__b",
        "assigned_child__grandchild_a__x",
        "assigned_child__grandchild_a__y",
        "assigned_child__grandchild_b__x",
        "assigned_child__grandchild_b__y",
        "fing",
    ]

    assert get_all_cols(Cols.assigned_child) == [
        "assigned_child__a",
        "assigned_child__b",
        "assigned_child__grandchild_a__x",
        "assigned_child__grandchild_a__y",
        "assigned_child__grandchild_b__x",
        "assigned_child__grandchild_b__y",
    ]


def test_accessor_id_cols():
    class IdCols(ColAccessor):

        fing_id = int
        other_id = str

    class TableCols(ColAccessor):

        foreign_key = IdCols.fing_id

    assert TableCols.foreign_key == "foreign_key"



import pandas as pd
import pytest

from colassigner import Col, ColAccessor, ColAssigner, get_all_cols


class Multi(ColAccessor):

    a = int
    b = float


def test_simple_annotation_assigner():
    class Simple(ColAccessor):

        ixx = str

    class Cas(ColAssigner):
        def works(self, df) -> Col[Simple]:
            return df.mean(axis=1)

        def with_str(self, df) -> Col[Multi.a]:
            return df.sum(axis=1)

        def no_effect(self, df) -> Col[int]:
            return df.sum(axis=1)

    assert Cas.works == "works__ixx"
    assert Cas.with_str == "with_str__a"
    assert get_all_cols(Cas) == ["no_effect", "with_str__a", "works__ixx"]

    df = pd.DataFrame({"a": [1, 2, 3], "b": [3, 2, 1]}).pipe(Cas())
    assert df.loc[:, Cas.works].tolist() == [2, 2, 2]
    assert df.loc[:, Cas.with_str].tolist() == [6, 6, 6]


def test_wrong_annotations():
    with pytest.raises(AssertionError):

        class Wrong(ColAssigner):
            def c1(self, _) -> int:
                pass  # pragma: no cover

        Wrong.c1

    with pytest.raises(ValueError):

        class Wrong2(ColAssigner):
            def thing(self, _) -> Col[Multi]:
                pass  # pragma: no cover

        Wrong2.thing



import ast
from importlib import import_module
from itertools import chain
from pathlib import Path
from typing import List, Type

from .core import ColAssigner


class MethodDef:
    def __init__(self, stmt: ast.FunctionDef, bases):
        self.name = stmt.name
        self.bases = set(bases)
        self.uses = set()
        self._adds(*stmt.body)

    def _add(self, elem: ast.AST):
        if isinstance(
            elem,
            (ast.Assign, ast.Return, ast.keyword, ast.Index, ast.Expr, ast.Starred),
        ):
            return self._add(elem.value)
        if isinstance(elem, (ast.List, ast.Tuple)):
            return self._adds(*elem.elts)
        if isinstance(elem, ast.If):
            return self._adds(elem.test, *elem.body, *elem.orelse)
        if isinstance(elem, ast.For):
            return self._adds(elem.iter, *elem.body)
        if isinstance(elem, ast.Try):
            return self._adds(*elem.handlers, *elem.body)
        if isinstance(elem, ast.ExceptHandler):
            return self._adds(*elem.body)
        if isinstance(elem, ast.Lambda):
            return self._add(elem.body)
        if isinstance(elem, ast.ExtSlice):
            return self._adds(*elem.dims)
        if isinstance(elem, ast.Slice):
            return self._adds(elem.lower, elem.upper)
        if isinstance(elem, ast.Attribute):
            base = elem.value
            if isinstance(base, ast.Name) and (base.id in [*self.bases, "self"]):
                return self.uses.add(elem.attr)
            return self._add(base)
        if isinstance(elem, ast.Call):
            self._adds(*elem.args, *elem.keywords)
            return self._add(elem.func)
        if isinstance(elem, ast.BinOp):
            return self._adds(elem.left, elem.right)
        if isinstance(elem, ast.Subscript):
            return self._adds(elem.value, elem.slice)
        if isinstance(elem, ast.Compare):
            return self._adds(elem.left, *elem.comparators)
        if (
            isinstance(elem, (ast.Constant, ast.Name, ast.Pass, ast.Raise))
            or elem is None
        ):
            return
        raise ValueError(
            f"unrecognized expression {type(elem)}: {elem}"
        )  # pragma: no cover

    def _adds(self, *args):
        return [*map(self._add, args)]


class ClsParser:
    def __init__(self, stmt: ast.ClassDef) -> None:
        self.name = stmt.name
        self._resolvers = {}
        self._mds: List[MethodDef] = []
        for fundef in stmt.body:
            md = MethodDef(fundef, [self.name])
            if md.name.startswith("_"):
                self._resolvers[md.name] = md.uses
            else:
                self._mds.append(md)

    def get_edges(self):
        return chain(*map(self._iter_mc, self._mds))

    def _iter_mc(self, md: MethodDef):
        for source in md.uses:
            if source.startswith("_"):
                for sub in self._resolve(source):
                    yield (sub, md.name)
            else:
                yield (source, md.name)

    def _resolve(self, source, resolved=()):
        for sub in self._resolvers.get(source, []):
            if sub in [source, *resolved]:
                continue
            resolved = (sub, *resolved)
            if sub.startswith("_"):
                for ssub in self._resolve(sub, resolved):
                    yield ssub
            else:
                yield sub


def get_dag(cls: Type[ColAssigner]):
    """generates a dag of the reliances of columns
    based on the ast of a colassigner

    Parameters
    ----------
    cls : Type[ColAssigner]

    Returns
    -------
    list of edges of a dag

    BETA! - WIP

    rules:
    - explicitly access columns within other functions
    - no external function referencing column (only method of assigner class)
    - no common source in as attributes
    - self should always be named self
    """
    fp = import_module(cls.__module__).__file__
    asm = ast.parse(Path(fp).read_text(), filename=fp)
    for stmt in asm.body:
        if isinstance(stmt, ast.ClassDef) and stmt.name == cls.__name__:
            return [*ClsParser(stmt).get_edges()]
